# -*- coding: utf-8 -*-
# Copyright (C) 2018-2020  Nexedi SA and Contributors.
#                          Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

"""Module wcfs.py provides python gateway for spawning and interoperating with wcfs server.

Join(zurl) joins wcfs server. If wcfs server for zurl is not yet running, it
will be automatically started if `autostart=True` is passed to join.
It will also be automatically started by default unless
$WENDELIN_CORE_WCFS_AUTOSTART=no is specified in environment.

WCFS represents filesystem-level connection to wcfs server. XXX created by join.
Conn represents logical connection that provides view of data on wcfs
filesystem as of particular database state.

WCFS represents connection to wcfs server obtained by join.
FileH ... XXX
XXX

Classes in wcfs module logically mirror classes in ZODB:

    wcfs.WCFS   <->  ZODB.DB
    wcfs.Conn   <->  ZODB.Connection


$WENDELIN_CORE_WCFS_AUTOSTART
$WENDELIN_CORE_WCFS_OPTIONS
"""

from __future__ import print_function, absolute_import

import os, sys, hashlib, tempfile, subprocess, time
import logging as log
from os.path import dirname
from errno import ENOENT, EEXIST

from golang import chan, select, default, func
from golang import sync, context
from golang.gcompat import qq

from persistent import Persistent
from ZODB.FileStorage import FileStorage
from zodbtools.util import ashex as h

from .client._wcfs import \
    PyWCFS          as _WCFS,       \
    PyWatchLink     as WatchLink,   \
    PyPinReq        as PinReq       \

# WCFS represents filesystem-level connection to wcfs server.
#
# Use join to create it.
#
# The primary way to access wcfs is to open logical connection viewing on-wcfs
# data as of particular database state, and use that logical connection to
# create base-layer mappings. See .connect and Conn for details.
#
# Raw files on wcfs can be accessed with ._path/._read/._stat/._open .
#
# WCFS logically mirrors ZODB.DB .
#
# XXX kill doc instead of C++.
class WCFS(_WCFS):
    # .mountpoint   path to wcfs mountpoint
    # ._fwcfs       /.wcfs/zurl opened to keep the server from going away (at least cleanly)
    # ._njoin       this connection was returned for so many joins

    # XXX for-testing only?
    # ._proc        wcfs process if it was opened by this WCFS | None
    pass



# ---- WCFS raw file access ----

# _path returns path for object on wcfs.
# - str:        wcfs root + obj;
# - Persistent: wcfs root + (head|@<at>)/bigfile/obj
@func(WCFS)
def _path(wc, obj, at=None):
    if isinstance(obj, Persistent):
        #assert type(obj) is ZBigFile   XXX import cycle
        objtypestr = type(obj).__module__ + "." + type(obj).__name__
        assert objtypestr == "wendelin.bigfile.file_zodb.ZBigFile", objtypestr
        head = "head/" if at is None else ("@%s/" % h(at))
        obj  = "%s/bigfile/%s" % (head, h(obj._p_oid))
        at   = None
    assert isinstance(obj, str)
    assert at is None  # must not be used with str
    return os.path.join(wc.mountpoint, obj)

# _read reads file corresponding to obj on wcfs.
@func(WCFS)
def _read(wc, obj, at=None):
    path = wc._path(obj, at=at)
    with open(path, 'rb') as f:     # XXX -> readfile
        return f.read()

# _stat stats file corresponding to obj on wcfs.
@func(WCFS)
def _stat(wc, obj, at=None):
    path = wc._path(obj, at=at)
    return os.stat(path)

# _open opens file corresponding to obj on wcfs.
@func(WCFS)
def _open(wc, obj, mode='rb', at=None):
    path = wc._path(obj, at=at)
    return open(path, mode, 0)  # unbuffered



# ---- join/run wcfs ----

_wcmu = sync.Mutex()
_wcregistry = {} # mntpt -> WCFS

@func(WCFS)
def __init__(wc, mountpoint, fwcfs, proc):
    wc.mountpoint = mountpoint
    wc._fwcfs     = fwcfs
    wc._njoin     = 1
    wc._proc      = proc

# close must be called to release joined connection after it is no longer needed.
@func(WCFS)
def close(wc):
    with _wcmu:
        wc._njoin -= 1
        if wc._njoin == 0:
            # XXX unmount wcfs as well?
            wc._fwcfs.close()
            del _wcregistry[wc.mountpoint]

# _default_autostart returns default autostart setting for join.
#
# Out-of-the-box we want wcfs to be automatically started, to ease developer
# experience when wendelin.core is standalone installed. However in environments
# like SlapOS, it is more preferable to start and monitor wcfs service explicitly.
# SlapOS & co. should thus set $WENDELIN_CORE_WCFS_AUTOSTART=no.
def _default_autostart():
    autostart = os.environ.get("WENDELIN_CORE_WCFS_AUTOSTART", "yes")
    autostart = autostart.lower()
    return {"yes": True, "no": False}[autostart]

# join connects to wcfs server for ZODB @ zurl.
#
# If wcfs for that zurl was already started, join connects to it.
# Otherwise it starts wcfs for zurl if autostart is True.
#
# For the same zurl join returns the same WCFS object.
def join(zurl, autostart=_default_autostart()): # -> WCFS
    mntpt = _mntpt_4zurl(zurl)
    with _wcmu:
        # check if we already have connection to wcfs server from this process
        wc = _wcregistry.get(mntpt)
        if wc is not None:
            wc._njoin += 1
            return wc

        # no. try opening .wcfs - if we succeed - wcfs is already mounted.
        try:
            f = open(mntpt + "/.wcfs/zurl")
        except IOError as e:
            if e.errno != ENOENT:
                raise
        else:
            # already have it
            wc = WCFS(mntpt, f, None)
            _wcregistry[mntpt] = wc
            return wc

        if not autostart:
            raise RuntimeError("wcfs: join %s: server not started" % zurl)

        # start wcfs with telling it to automatically exit when there is no client activity.
        optv_extra = os.environ.get("WENDELIN_CORE_WCFS_OPTIONS", "").split()
        return _start(zurl, "-autoexit", *optv_extra)


# _start starts wcfs server for ZODB @ zurl.
#
# optv can be optionally given to pass flags to wcfs.
# called under _wcmu.
def _start(zurl, *optv):    # -> WCFS
    mntpt = _mntpt_4zurl(zurl)
    log.info("wcfs: starting for %s ...", zurl)

    # XXX errctx "wcfs: start"

    # spawn wcfs and wait till filesystem-level access to it is ready
    wc = WCFS(mntpt, None, None)
    wg = sync.WorkGroup(context.background())
    fsready = chan(dtype='C.structZ')
    def _(ctx):
        # XXX errctx "spawn"
        argv = [_wcfs_exe()] + list(optv) + [zurl, mntpt]
        proc = subprocess.Popen(argv, close_fds=True)
        while 1:
            ret = proc.poll()
            if ret is not None:
                raise RuntimeError("exited with %s" % ret)

            _, _rx = select(
                ctx.done().recv,    # 0
                fsready.recv,       # 1
                default,            # 2
            )
            if _ == 0:
                proc.terminate()
                raise ctx.err()
            if _ == 1:
                # startup was ok - don't monitor spawned wcfs any longer
                wc._proc = proc
                return

            time.sleep(0.1)
    wg.go(_)

    def _(ctx):
        # XXX errctx "waitmount"
        while 1:
            try:
                f = open("%s/.wcfs/zurl" % mntpt)
            except IOError as e:
                if e.errno != ENOENT:
                    raise
            else:
                dotwcfs = f.read()
                if dotwcfs != zurl:
                    raise RuntimeError(".wcfs/zurl != zurl  (%s != %s)" % (qq(dotwcfs), qq(zurl)))

                wc._fwcfs = f
                fsready.close()
                return

            _, _rx = select(
                ctx.done().recv,    # 0
                default,            # 1
            )
            if _ == 0:
                raise ctx.err()

            time.sleep(0.1)
    wg.go(_)

    wg.wait()

    assert mntpt not in _wcregistry
    _wcregistry[mntpt] = wc
    return wc


# ---- misc ----

# _wcfs_exe returns path to wcfs executable.
def _wcfs_exe():
    return '%s/wcfs' % dirname(__file__)

# _mntpt_4zurl returns wcfs should-be mountpoint for ZODB @ zurl.
#
# it also makes sure the mountpoint exists.
def _mntpt_4zurl(zurl):
    # XXX what if zurl is zconfig://... ? -> then we have to look inside?
    #     -> zstor_2zurl extracts zurl in canonical form and zconfig:// is not possible there.
    m = hashlib.sha1()
    m.update(zurl)
    mntpt = "%s/wcfs/%s" % (tempfile.gettempdir(), m.hexdigest())
    _mkdir_p(mntpt)
    return mntpt


# zstor_2zurl converts a ZODB storage to URL to access it.
# XXX -> unexport?
def zstor_2zurl(zstor):
    # There is, sadly, no unified way to do it, as even if storages are created via
    # zodburi, after creation its uri is lost. And storages could be created not
    # only through URI but e.g. via ZConfig and manually. We want to support all
    # those cases...
    #
    # For this reason extract URL with important for wcfs use-case parameters in
    # ad-hoc way.
    if isinstance(zstor, FileStorage):
        return "file://%s" % (zstor._file_name,)

    # TODO ZEO + NEO support
    raise NotImplementedError("don't know how to extract zurl from %r" % zstor)


# mkdir -p.
def _mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != EEXIST:
            raise


# serve starts and runs wcfs server for ZODB @ zurl.
#
# it mounts wcfs at a location that is with 1-1 correspondence with zurl.
# it then waits for wcfs to exit (either due to unmount or an error).
#
# it is an error if wcfs was already started.
#
# XXX optv
# if exec_ is True, wcfs is not spawned, but executed into.
#
# serve(zurl, exec_=False).
def serve(zurl, optv, exec_=False):
    mntpt = _mntpt_4zurl(zurl)

    # try opening .wcfs - it is an error if we can do it.
    # XXX -> option to wcfs itself?
    try:
        f = open(mntpt + "/.wcfs/zurl")
    except IOError as e:
        if e.errno != ENOENT:
            raise
    else:
        f.close()
        raise RuntimeError("wcfs: start %s: already started" % zurl)

    # seems to be ok to start
    # XXX race window if something starts after ^^^ check
    argv = [_wcfs_exe()] + list(optv) + [zurl, mntpt]
    if not exec_:
        subprocess.check_call(argv, close_fds=True)
    else:
        os.execv(argv[0], argv)




# if called as main just -> serve()
def main():
    argv = sys.argv[1:]
    # XXX usage
    zurl = argv[-1]     # -a -b zurl    -> zurl
    optv = argv[:-1]    # -a -b zurl    -> -a -b
    serve(zurl, optv, exec_=True)
